@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2004-2012  Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (C) 2004-2016  Internet Systems Consortium, Inc. ("ISC")
  * Copyright (C) 1999-2003  Internet Software Consortium.
  *
  * Permission to use, copy, modify, and/or distribute this software for any
@@ -15,8 +15,6 @@
  * PERFORMANCE OF THIS SOFTWARE.
  */
 
-/* $Id$ */
-
 #include <config.h>
 
 #include <isc/formatcheck.h>
@@ -25,6 +23,9 @@
 #include <isc/platform.h>
 #include <isc/print.h>
 #include <isc/queue.h>
+#include <isc/random.h>
+#include <isc/safe.h>
+#include <isc/serial.h>
 #include <isc/stats.h>
 #include <isc/stdio.h>
 #include <isc/string.h>
@@ -32,6 +33,12 @@
 #include <isc/timer.h>
 #include <isc/util.h>
 
+#ifdef AES_SIT
+#include <isc/aes.h>
+#else
+#include <isc/hmacsha.h>
+#endif
+
 #include <dns/db.h>
 #include <dns/dispatch.h>
 #include <dns/events.h>
@@ -113,6 +120,11 @@
  */
 #endif
 
+#define COOKIE_SIZE 24U /* 8 + 4 + 4 + 8 */
+
+#define WANTNSID(x) (((x)->attributes & NS_CLIENTATTR_WANTNSID) != 0)
+#define WANTEXPIRE(x) (((x)->attributes & NS_CLIENTATTR_WANTEXPIRE) != 0)
+
 /*% nameserver client manager structure */
 struct ns_clientmgr {
 	/* Unlocked. */
@@ -233,6 +245,12 @@ static void client_request(isc_task_t *task, isc_event_t *event);
 static void ns_client_dumpmessage(ns_client_t *client, const char *reason);
 static isc_result_t get_client(ns_clientmgr_t *manager, ns_interface_t *ifp,
 			       dns_dispatch_t *disp, isc_boolean_t tcp);
+static inline isc_boolean_t
+allowed(isc_netaddr_t *addr, dns_name_t *signer, dns_acl_t *acl);
+#ifdef ISC_PLATFORM_USESIT
+static void compute_cookie(ns_client_t *client, isc_uint32_t when,
+			isc_uint32_t nonce, isc_buffer_t *buf);
+#endif
 
 void
 ns_client_recursing(ns_client_t *client) {
@@ -326,12 +344,12 @@ exit_check(ns_client_t *client) {
 		 * We are trying to abort request processing.
 		 */
 		if (client->nsends > 0) {
-			isc_socket_t *socket;
+			isc_socket_t *sock;
 			if (TCP_CLIENT(client))
-				socket = client->tcpsocket;
+				sock = client->tcpsocket;
 			else
-				socket = client->udpsocket;
-			isc_socket_cancel(socket, client->task,
+				sock = client->udpsocket;
+			isc_socket_cancel(sock, client->task,
 					  ISC_SOCKCANCEL_SEND);
 		}
 
@@ -501,12 +519,14 @@ exit_check(ns_client_t *client) {
 		 * the dying client inbetween.
 		 */
 		client->state = NS_CLIENTSTATE_INACTIVE;
-		if (!ns_g_clienttest)
-			ISC_QUEUE_PUSH(manager->inactive, client, ilink);
 		INSIST(client->recursionquota == NULL);
 
 		if (client->state == client->newstate) {
 			client->newstate = NS_CLIENTSTATE_MAX;
+			if (!ns_g_clienttest && manager != NULL &&
+			    !manager->exiting)
+				ISC_QUEUE_PUSH(manager->inactive, client,
+					       ilink);
 			if (client->needshutdown)
 				isc_task_shutdown(client->task);
 			return (ISC_TRUE);
@@ -526,12 +546,26 @@ exit_check(ns_client_t *client) {
 		REQUIRE(client->state == NS_CLIENTSTATE_INACTIVE);
 
 		INSIST(client->recursionquota == NULL);
+		INSIST(!ISC_QLINK_LINKED(client, ilink));
+
+		if (manager != NULL) {
+			LOCK(&manager->listlock);
+			ISC_LIST_UNLINK(manager->clients, client, link);
+			LOCK(&manager->lock);
+			if (manager->exiting &&
+			    ISC_LIST_EMPTY(manager->clients))
+				destroy_manager = ISC_TRUE;
+			UNLOCK(&manager->lock);
+			UNLOCK(&manager->listlock);
+		}
 
 		ns_query_free(client);
 		isc_mem_put(client->mctx, client->recvbuf, RECV_BUFFER_SIZE);
 		isc_event_free((isc_event_t **)&client->sendevent);
 		isc_event_free((isc_event_t **)&client->recvevent);
 		isc_timer_detach(&client->timer);
+		if (client->delaytimer != NULL)
+			isc_timer_detach(&client->delaytimer);
 
 		if (client->tcpbuf != NULL)
 			isc_mem_put(client->mctx, client->tcpbuf,
@@ -544,16 +578,6 @@ exit_check(ns_client_t *client) {
 		}
 
 		dns_message_destroy(&client->message);
-		if (manager != NULL) {
-			LOCK(&manager->listlock);
-			ISC_LIST_UNLINK(manager->clients, client, link);
-			LOCK(&manager->lock);
-			if (manager->exiting &&
-			    ISC_LIST_EMPTY(manager->clients))
-				destroy_manager = ISC_TRUE;
-			UNLOCK(&manager->lock);
-			UNLOCK(&manager->listlock);
-		}
 
 		/*
 		 * Detaching the task must be done after unlinking from
@@ -574,6 +598,13 @@ exit_check(ns_client_t *client) {
 			isc_mem_stats(client->mctx, stderr);
 			INSIST(0);
 		}
+
+		/*
+		 * Destroy the fetchlock mutex that was created in
+		 * ns_query_init().
+		 */
+		DESTROYLOCK(&client->query.fetchlock);
+
 		isc_mem_putanddetach(&client->mctx, client, sizeof(*client));
 	}
 
@@ -634,6 +665,9 @@ client_shutdown(isc_task_t *task, isc_event_t *event) {
 		client->shutdown_arg = NULL;
 	}
 
+	if (ISC_QLINK_LINKED(client, ilink))
+		ISC_QUEUE_UNLINK(client->manager->inactive, client, ilink);
+
 	client->newstate = NS_CLIENTSTATE_FREED;
 	client->needshutdown = ISC_FALSE;
 	(void)exit_check(client);
@@ -670,8 +704,11 @@ ns_client_endrequest(ns_client_t *client) {
 	client->ednsversion = -1;
 	dns_message_reset(client->message, DNS_MESSAGE_INTENTPARSE);
 
-	if (client->recursionquota != NULL)
+	if (client->recursionquota != NULL) {
 		isc_quota_detach(&client->recursionquota);
+		isc_stats_decrement(ns_g_server->nsstats,
+				    dns_nsstatscounter_recursclients);
+	}
 
 	/*
 	 * Clear all client attributes that are specific to
@@ -789,10 +826,24 @@ client_allocsendbuf(ns_client_t *client, isc_buffer_t *buffer,
 		}
 	} else {
 		data = sendbuf;
+#ifdef ISC_PLATFORM_USESIT
+		if ((client->attributes & NS_CLIENTATTR_HAVESIT) == 0) {
+			if (client->view != NULL)
+				bufsize = client->view->situdp;
+			else
+				bufsize = 512;
+		} else
+			bufsize = client->udpsize;
+		if (bufsize > client->udpsize)
+			bufsize = client->udpsize;
+		if (bufsize > SEND_BUFFER_SIZE)
+			bufsize = SEND_BUFFER_SIZE;
+#else
 		if (client->udpsize < SEND_BUFFER_SIZE)
 			bufsize = client->udpsize;
 		else
 			bufsize = SEND_BUFFER_SIZE;
+#endif
 		if (length > bufsize) {
 			result = ISC_R_NOSPACE;
 			goto done;
@@ -812,16 +863,17 @@ client_sendpkg(ns_client_t *client, isc_buffer_t *buffer) {
 	isc_result_t result;
 	isc_region_t r;
 	isc_sockaddr_t *address;
-	isc_socket_t *socket;
+	isc_socket_t *sock;
 	isc_netaddr_t netaddr;
 	int match;
 	unsigned int sockflags = ISC_SOCKFLAG_IMMEDIATE;
+	isc_dscp_t dispdscp = -1;
 
 	if (TCP_CLIENT(client)) {
-		socket = client->tcpsocket;
+		sock = client->tcpsocket;
 		address = NULL;
 	} else {
-		socket = client->udpsocket;
+		sock = client->udpsocket;
 		address = &client->peeraddr;
 
 		isc_netaddr_fromsockaddr(&netaddr, &client->peeraddr);
@@ -841,11 +893,25 @@ client_sendpkg(ns_client_t *client, isc_buffer_t *buffer) {
 	else
 		pktinfo = NULL;
 
+	if (client->dispatch != NULL) {
+		dispdscp = dns_dispatch_getdscp(client->dispatch);
+		if (dispdscp != -1)
+			client->dscp = dispdscp;
+	}
+
+	if (client->dscp == -1) {
+		client->sendevent->attributes &= ~ISC_SOCKEVENTATTR_DSCP;
+		client->sendevent->dscp = 0;
+	} else {
+		client->sendevent->attributes |= ISC_SOCKEVENTATTR_DSCP;
+		client->sendevent->dscp = client->dscp;
+	}
+
 	isc_buffer_usedregion(buffer, &r);
 
 	CTRACE("sendto");
 
-	result = isc_socket_sendto2(socket, &r, client->task,
+	result = isc_socket_sendto2(sock, &r, client->task,
 				    address, pktinfo,
 				    client->sendevent, sockflags);
 	if (result == ISC_R_SUCCESS || result == ISC_R_INPROGRESS) {
@@ -904,8 +970,8 @@ ns_client_sendraw(ns_client_t *client, dns_message_t *message) {
 	ns_client_next(client, result);
 }
 
-void
-ns_client_send(ns_client_t *client) {
+static void
+client_send(ns_client_t *client) {
 	isc_result_t result;
 	unsigned char *data;
 	isc_buffer_t buffer;
@@ -937,11 +1003,18 @@ ns_client_send(ns_client_t *client) {
 		else if (client->view->preferred_glue == dns_rdatatype_aaaa)
 			preferred_glue = DNS_MESSAGERENDER_PREFER_AAAA;
 	}
+	if (preferred_glue == 0) {
+		if (isc_sockaddr_pf(&client->peeraddr) == AF_INET)
+			preferred_glue = DNS_MESSAGERENDER_PREFER_A;
+		else
+			preferred_glue = DNS_MESSAGERENDER_PREFER_AAAA;
+	}
 
-#ifdef ALLOW_FILTER_AAAA_ON_V4
+#ifdef ALLOW_FILTER_AAAA
 	/*
 	 * filter-aaaa-on-v4 yes or break-dnssec option to suppress
-	 * AAAA records
+	 * AAAA records.
+	 *
 	 * We already know that request came via IPv4,
 	 * that we have both AAAA and A records,
 	 * and that we either have no signatures that the client wants
@@ -956,6 +1029,16 @@ ns_client_send(ns_client_t *client) {
 	}
 #endif
 
+	/*
+	 * Create an OPT for our reply.
+	 */
+	if ((client->attributes & NS_CLIENTATTR_WANTOPT) != 0) {
+		result = ns_client_addopt(client, client->message,
+					  &client->opt);
+		if (result != ISC_R_SUCCESS)
+			goto done;
+	}
+
 	/*
 	 * XXXRTH  The following doesn't deal with TCP buffer resizing.
 	 */
@@ -967,6 +1050,19 @@ ns_client_send(ns_client_t *client) {
 	result = dns_compress_init(&cctx, -1, client->mctx);
 	if (result != ISC_R_SUCCESS)
 		goto done;
+	if (client->peeraddr_valid && client->view != NULL) {
+		isc_netaddr_t netaddr;
+		dns_name_t *name = NULL;
+
+		isc_netaddr_fromsockaddr(&netaddr, &client->peeraddr);
+		if (client->message->tsigkey != NULL)
+			name = &client->message->tsigkey->name;
+		if (client->view->nocasecompress == NULL ||
+		    !allowed(&netaddr, name, client->view->nocasecompress))
+		{
+			dns_compress_setsensitive(&cctx, ISC_TRUE);
+		}
+	}
 	cleanup_cctx = ISC_TRUE;
 
 	result = dns_message_renderbegin(client->message, &cctx, &buffer);
@@ -988,6 +1084,11 @@ ns_client_send(ns_client_t *client) {
 	}
 	if (result != ISC_R_SUCCESS)
 		goto done;
+	/*
+	 * Stop after the question if TC was set for rate limiting.
+	 */
+	if ((client->message->flags & DNS_MESSAGEFLAG_TC) != 0)
+		goto renderend;
 	result = dns_message_rendersection(client->message,
 					   DNS_SECTION_ANSWER,
 					   DNS_MESSAGERENDER_PARTIAL |
@@ -1065,6 +1166,72 @@ ns_client_send(ns_client_t *client) {
 	ns_client_next(client, result);
 }
 
+/*
+ * Completes the sending of a delayed client response.
+ */
+static void
+client_delay(isc_task_t *task, isc_event_t *event) {
+	ns_client_t *client;
+
+	REQUIRE(event != NULL);
+	REQUIRE(event->ev_type == ISC_TIMEREVENT_LIFE ||
+		event->ev_type == ISC_TIMEREVENT_IDLE);
+	client = event->ev_arg;
+	REQUIRE(NS_CLIENT_VALID(client));
+	REQUIRE(task == client->task);
+	REQUIRE(client->delaytimer != NULL);
+
+	UNUSED(task);
+
+	CTRACE("client_delay");
+
+	isc_event_free(&event);
+	isc_timer_detach(&client->delaytimer);
+
+	client_send(client);
+	ns_client_detach(&client);
+}
+
+void
+ns_client_send(ns_client_t *client) {
+
+	/*
+	 * Delay the response by ns_g_delay ms.
+	 */
+	if (ns_g_delay != 0) {
+		ns_client_t *dummy = NULL;
+		isc_result_t result;
+		isc_interval_t interval;
+
+		/*
+		 * Replace ourselves if we have not already been replaced.
+		 */
+		if (!client->mortal) {
+			result = ns_client_replace(client);
+			if (result != ISC_R_SUCCESS)
+				goto nodelay;
+		}
+
+		ns_client_attach(client, &dummy);
+		if (ns_g_delay >= 1000)
+			isc_interval_set(&interval, ns_g_delay / 1000,
+					 (ns_g_delay % 1000) * 1000000);
+		else
+			isc_interval_set(&interval, 0, ns_g_delay * 1000000);
+		result = isc_timer_create(client->manager->timermgr,
+					  isc_timertype_once, NULL, &interval,
+					  client->task, client_delay,
+					  client, &client->delaytimer);
+		if (result == ISC_R_SUCCESS)
+			return;
+
+		ns_client_detach(&dummy);
+	}
+
+ nodelay:
+	client_send(client);
+}
+
 #if NS_CLIENT_DROPPORT
 #define DROPPORT_NO		0
 #define DROPPORT_REQUEST	1
@@ -1127,6 +1294,56 @@ ns_client_error(ns_client_t *client, isc_result_t result) {
 	}
 #endif
 
+	/*
+	 * Try to rate limit error responses.
+	 */
+	if (client->view != NULL && client->view->rrl != NULL) {
+		isc_boolean_t wouldlog;
+		char log_buf[DNS_RRL_LOG_BUF_LEN];
+		dns_rrl_result_t rrl_result;
+		int loglevel;
+
+		INSIST(rcode != dns_rcode_noerror &&
+		       rcode != dns_rcode_nxdomain);
+		if (ns_g_server->log_queries)
+			loglevel = DNS_RRL_LOG_DROP;
+		else
+			loglevel = ISC_LOG_DEBUG(1);
+		wouldlog = isc_log_wouldlog(ns_g_lctx, loglevel);
+		rrl_result = dns_rrl(client->view, &client->peeraddr,
+				     TCP_CLIENT(client),
+				     dns_rdataclass_in, dns_rdatatype_none,
+				     NULL, result, client->now,
+				     wouldlog, log_buf, sizeof(log_buf));
+		if (rrl_result != DNS_RRL_RESULT_OK) {
+			/*
+			 * Log dropped errors in the query category
+			 * so that they are not lost in silence.
+			 * Starts of rate-limited bursts are logged in
+			 * NS_LOGCATEGORY_RRL.
+			 */
+			if (wouldlog) {
+				ns_client_log(client,
+					      NS_LOGCATEGORY_QUERY_ERRORS,
+					      NS_LOGMODULE_CLIENT,
+					      loglevel,
+					      "%s", log_buf);
+			}
+			/*
+			 * Some error responses cannot be 'slipped',
+			 * so don't try to slip any error responses.
+			 */
+			if (!client->view->rrl->log_only) {
+				isc_stats_increment(ns_g_server->nsstats,
+						dns_nsstatscounter_ratedropped);
+				isc_stats_increment(ns_g_server->nsstats,
+						dns_nsstatscounter_dropped);
+				ns_client_next(client, DNS_R_DROP);
+				return;
+			}
+		}
+	}
+
 	/*
 	 * Message may be an in-progress reply that we had trouble
 	 * with, in which case QR will be set.  We need to clear QR before
@@ -1181,66 +1398,41 @@ ns_client_error(ns_client_t *client, isc_result_t result) {
 	ns_client_send(client);
 }
 
-static inline isc_result_t
-client_addopt(ns_client_t *client) {
-	dns_rdataset_t *rdataset;
-	dns_rdatalist_t *rdatalist;
-	dns_rdata_t *rdata;
+isc_result_t
+ns_client_addopt(ns_client_t *client, dns_message_t *message,
+		 dns_rdataset_t **opt)
+{
+	char nsid[BUFSIZ], *nsidp;
+#ifdef ISC_PLATFORM_USESIT
+	unsigned char cookie[COOKIE_SIZE];
+#endif
 	isc_result_t result;
 	dns_view_t *view;
 	dns_resolver_t *resolver;
 	isc_uint16_t udpsize;
+	dns_ednsopt_t ednsopts[DNS_EDNSOPTIONS];
+	int count = 0;
+	unsigned int flags;
+	unsigned char expire[4];
 
-	REQUIRE(client->opt == NULL);	/* XXXRTH free old. */
+	REQUIRE(NS_CLIENT_VALID(client));
+	REQUIRE(opt != NULL && *opt == NULL);
+	REQUIRE(message != NULL);
 
-	rdatalist = NULL;
-	result = dns_message_gettemprdatalist(client->message, &rdatalist);
-	if (result != ISC_R_SUCCESS)
-		return (result);
-	rdata = NULL;
-	result = dns_message_gettemprdata(client->message, &rdata);
-	if (result != ISC_R_SUCCESS)
-		return (result);
-	rdataset = NULL;
-	result = dns_message_gettemprdataset(client->message, &rdataset);
-	if (result != ISC_R_SUCCESS)
-		return (result);
-	dns_rdataset_init(rdataset);
-
-	rdatalist->type = dns_rdatatype_opt;
-	rdatalist->covers = 0;
-
-	/*
-	 * Set the maximum UDP buffer size.
-	 */
 	view = client->view;
 	resolver = (view != NULL) ? view->resolver : NULL;
 	if (resolver != NULL)
 		udpsize = dns_resolver_getudpsize(resolver);
 	else
 		udpsize = ns_g_udpsize;
-	rdatalist->rdclass = udpsize;
 
-	/*
-	 * Set EXTENDED-RCODE, VERSION and Z to 0.
-	 */
-	rdatalist->ttl = (client->extflags & DNS_MESSAGEEXTFLAG_REPLYPRESERVE);
+	flags = client->extflags & DNS_MESSAGEEXTFLAG_REPLYPRESERVE;
 
 	/* Set EDNS options if applicable */
-	if (client->attributes & NS_CLIENTATTR_WANTNSID &&
+	if (WANTNSID(client) &&
 	    (ns_g_server->server_id != NULL ||
 	     ns_g_server->server_usehostname)) {
-		/*
-		 * Space required for NSID data:
-		 *   2 bytes for opt code
-		 * + 2 bytes for NSID length
-		 * + NSID itself
-		 */
-		char nsid[BUFSIZ], *nsidp;
-		isc_buffer_t *buffer = NULL;
-
 		if (ns_g_server->server_usehostname) {
-			isc_result_t result;
 			result = ns_os_gethostname(nsid, sizeof(nsid));
 			if (result != ISC_R_SUCCESS) {
 				goto no_nsid;
@@ -1249,35 +1441,48 @@ client_addopt(ns_client_t *client) {
 		} else
 			nsidp = ns_g_server->server_id;
 
-		rdata->length = strlen(nsidp) + 4;
-		result = isc_buffer_allocate(client->mctx, &buffer,
-					     rdata->length);
-		if (result != ISC_R_SUCCESS)
-			goto no_nsid;
+		INSIST(count < DNS_EDNSOPTIONS);
+		ednsopts[count].code = DNS_OPT_NSID;
+		ednsopts[count].length = strlen(nsidp);
+		ednsopts[count].value = (unsigned char *)nsidp;
+		count++;
+	}
+ no_nsid:
+#ifdef ISC_PLATFORM_USESIT
+	if ((client->attributes & NS_CLIENTATTR_WANTSIT) != 0) {
+		isc_buffer_t buf;
+		isc_stdtime_t now;
+		isc_uint32_t nonce;
 
-		isc_buffer_putuint16(buffer, DNS_OPT_NSID);
-		isc_buffer_putuint16(buffer, strlen(nsidp));
-		isc_buffer_putstr(buffer, nsidp);
-		rdata->data = buffer->base;
-		dns_message_takebuffer(client->message, &buffer);
-	} else {
-no_nsid:
-		rdata->data = NULL;
-		rdata->length = 0;
+		isc_buffer_init(&buf, cookie, sizeof(cookie));
+		isc_stdtime_get(&now);
+		isc_random_get(&nonce);
+
+		compute_cookie(client, now, nonce, &buf);
+
+		INSIST(count < DNS_EDNSOPTIONS);
+		ednsopts[count].code = DNS_OPT_COOKIE;
+		ednsopts[count].length = COOKIE_SIZE;
+		ednsopts[count].value = cookie;
+		count++;
+	}
+#endif
+	if ((client->attributes & NS_CLIENTATTR_HAVEEXPIRE) != 0) {
+		isc_buffer_t buf;
+
+		INSIST(count < DNS_EDNSOPTIONS);
+
+		isc_buffer_init(&buf, expire, sizeof(expire));
+		isc_buffer_putuint32(&buf, client->expire);
+		ednsopts[count].code = DNS_OPT_EXPIRE;
+		ednsopts[count].length = 4;
+		ednsopts[count].value = expire;
+		count++;
 	}
 
-	rdata->rdclass = rdatalist->rdclass;
-	rdata->type = rdatalist->type;
-	rdata->flags = 0;
-
-	ISC_LIST_INIT(rdatalist->rdata);
-	ISC_LIST_APPEND(rdatalist->rdata, rdata, link);
-	RUNTIME_CHECK(dns_rdatalist_tordataset(rdatalist, rdataset)
-		      == ISC_R_SUCCESS);
-
-	client->opt = rdataset;
-
-	return (ISC_R_SUCCESS);
+	result = dns_message_buildopt(message, opt, 0, udpsize, flags,
+				      ednsopts, count);
+	return (result);
 }
 
 static inline isc_boolean_t
@@ -1358,6 +1563,277 @@ ns_client_isself(dns_view_t *myview, dns_tsigkey_t *mykey,
 	return (ISC_TF(view == myview));
 }
 
+#ifdef ISC_PLATFORM_USESIT
+static void
+compute_cookie(ns_client_t *client, isc_uint32_t when, isc_uint32_t nonce,
+	       isc_buffer_t *buf)
+{
+#ifdef AES_SIT
+	unsigned char digest[ISC_AES_BLOCK_LENGTH];
+	unsigned char input[4 + 4 + 16];
+	isc_netaddr_t netaddr;
+	unsigned char *cp;
+	unsigned int i;
+
+	memset(input, 0, sizeof(input));
+	cp = isc_buffer_used(buf);
+	isc_buffer_putmem(buf, client->cookie, 8);
+	isc_buffer_putuint32(buf, nonce);
+	isc_buffer_putuint32(buf, when);
+	memmove(input, cp, 16);
+	isc_aes128_crypt(ns_g_server->secret, input, digest);
+	for (i = 0; i < 8; i++)
+		input[i] = digest[i] ^ digest[i + 8];
+	isc_netaddr_fromsockaddr(&netaddr, &client->peeraddr);
+	switch (netaddr.family) {
+	case AF_INET:
+		memmove(input + 8, (unsigned char *)&netaddr.type.in, 4);
+		memset(input + 12, 0, 4);
+		isc_aes128_crypt(ns_g_server->secret, input, digest);
+		break;
+	case AF_INET6:
+		memmove(input + 8, (unsigned char *)&netaddr.type.in6, 16);
+		isc_aes128_crypt(ns_g_server->secret, input, digest);
+		for (i = 0; i < 8; i++)
+			input[i + 8] = digest[i] ^ digest[i + 8];
+		isc_aes128_crypt(ns_g_server->secret, input + 8, digest);
+		break;
+	}
+	for (i = 0; i < 8; i++)
+		digest[i] ^= digest[i + 8];
+	isc_buffer_putmem(buf, digest, 8);
+#endif
+#ifdef HMAC_SHA1_SIT
+	unsigned char digest[ISC_SHA1_DIGESTLENGTH];
+	isc_netaddr_t netaddr;
+	unsigned char *cp;
+	isc_hmacsha1_t hmacsha1;
+
+	cp = isc_buffer_used(buf);
+	isc_buffer_putmem(buf, client->cookie, 8);
+	isc_buffer_putuint32(buf, nonce);
+	isc_buffer_putuint32(buf, when);
+
+	isc_hmacsha1_init(&hmacsha1,
+			  ns_g_server->secret,
+			  ISC_SHA1_DIGESTLENGTH);
+	isc_hmacsha1_update(&hmacsha1, cp, 16);
+	isc_netaddr_fromsockaddr(&netaddr, &client->peeraddr);
+	switch (netaddr.family) {
+	case AF_INET:
+		isc_hmacsha1_update(&hmacsha1,
+				    (unsigned char *)&netaddr.type.in, 4);
+		break;
+	case AF_INET6:
+		isc_hmacsha1_update(&hmacsha1,
+				    (unsigned char *)&netaddr.type.in6, 16);
+		break;
+	}
+	isc_hmacsha1_update(&hmacsha1, client->cookie, sizeof(client->cookie));
+	isc_hmacsha1_sign(&hmacsha1, digest, sizeof(digest));
+	isc_buffer_putmem(buf, digest, 8);
+	isc_hmacsha1_invalidate(&hmacsha1);
+#endif
+#ifdef HMAC_SHA256_SIT
+	unsigned char digest[ISC_SHA256_DIGESTLENGTH];
+	isc_netaddr_t netaddr;
+	unsigned char *cp;
+	isc_hmacsha256_t hmacsha256;
+
+	cp = isc_buffer_used(buf);
+	isc_buffer_putmem(buf, client->cookie, 8);
+	isc_buffer_putuint32(buf, nonce);
+	isc_buffer_putuint32(buf, when);
+
+	isc_hmacsha256_init(&hmacsha256,
+			    ns_g_server->secret,
+			    ISC_SHA256_DIGESTLENGTH);
+	isc_hmacsha256_update(&hmacsha256, cp, 16);
+	isc_netaddr_fromsockaddr(&netaddr, &client->peeraddr);
+	switch (netaddr.family) {
+	case AF_INET:
+		isc_hmacsha256_update(&hmacsha256,
+				      (unsigned char *)&netaddr.type.in, 4);
+		break;
+	case AF_INET6:
+		isc_hmacsha256_update(&hmacsha256,
+				      (unsigned char *)&netaddr.type.in6, 16);
+		break;
+	}
+	isc_hmacsha256_update(&hmacsha256, client->cookie,
+			      sizeof(client->cookie));
+	isc_hmacsha256_sign(&hmacsha256, digest, sizeof(digest));
+	isc_buffer_putmem(buf, digest, 8);
+	isc_hmacsha256_invalidate(&hmacsha256);
+#endif
+}
+
+static void
+process_cookie(ns_client_t *client, isc_buffer_t *buf, size_t optlen) {
+	unsigned char dbuf[COOKIE_SIZE];
+	unsigned char *old;
+	isc_stdtime_t now;
+	isc_uint32_t when;
+	isc_uint32_t nonce;
+	isc_buffer_t db;
+
+	/*
+	 * If we have already seen a SIT option skip this SIT option.
+	 */
+	if ((client->attributes & NS_CLIENTATTR_WANTSIT) != 0) {
+		isc_buffer_forward(buf, (isc_uint32_t)optlen);
+		return;
+	}
+	client->attributes |= NS_CLIENTATTR_WANTSIT;
+
+	isc_stats_increment(ns_g_server->nsstats,
+			    dns_nsstatscounter_sitopt);
+
+	if (optlen != COOKIE_SIZE) {
+		/*
+		 * Not our token.
+		 */
+		if (optlen >= 8U)
+			memmove(client->cookie, isc_buffer_current(buf), 8);
+		else
+			memset(client->cookie, 0, 8);
+		isc_buffer_forward(buf, (unsigned int)optlen);
+
+		if (optlen == 8U)
+			isc_stats_increment(ns_g_server->nsstats,
+					    dns_nsstatscounter_sitnew);
+		else
+			isc_stats_increment(ns_g_server->nsstats,
+					    dns_nsstatscounter_sitbadsize);
+		return;
+	}
+
+	/*
+	 * Process all of the incoming buffer.
+	 */
+	old = isc_buffer_current(buf);
+	memmove(client->cookie, old, 8);
+	isc_buffer_forward(buf, 8);
+	nonce = isc_buffer_getuint32(buf);
+	when = isc_buffer_getuint32(buf);
+	isc_buffer_forward(buf, 8);
+
+	/*
+	 * Allow for a 5 minute clock skew between servers sharing a secret.
+	 * Only accept COOKIE if we have talked to the client in the last hour.
+	 */
+	isc_stdtime_get(&now);
+	if (isc_serial_gt(when, (now + 300)) ||		/* In the future. */
+	    isc_serial_lt(when, (now - 3600))) {	/* In the past. */
+		isc_stats_increment(ns_g_server->nsstats,
+				    dns_nsstatscounter_sitbadtime);
+		return;
+	}
+
+	isc_buffer_init(&db, dbuf, sizeof(dbuf));
+	compute_cookie(client, when, nonce, &db);
+
+	if (!isc_safe_memequal(old, dbuf, COOKIE_SIZE)) {
+		isc_stats_increment(ns_g_server->nsstats,
+				    dns_nsstatscounter_sitnomatch);
+		return;
+	}
+	isc_stats_increment(ns_g_server->nsstats,
+			    dns_nsstatscounter_sitmatch);
+	client->attributes |= NS_CLIENTATTR_HAVESIT;
+}
+#endif
+
+static isc_result_t
+process_opt(ns_client_t *client, dns_rdataset_t *opt) {
+	dns_rdata_t rdata;
+	isc_buffer_t optbuf;
+	isc_result_t result;
+	isc_uint16_t optcode;
+	isc_uint16_t optlen;
+
+	/*
+	 * Set the client's UDP buffer size.
+	 */
+	client->udpsize = opt->rdclass;
+
+	/*
+	 * If the requested UDP buffer size is less than 512,
+	 * ignore it and use 512.
+	 */
+	if (client->udpsize < 512)
+		client->udpsize = 512;
+
+	/*
+	 * Get the flags out of the OPT record.
+	 */
+	client->extflags = (isc_uint16_t)(opt->ttl & 0xFFFF);
+
+	/*
+	 * Do we understand this version of EDNS?
+	 *
+	 * XXXRTH need library support for this!
+	 */
+	client->ednsversion = (opt->ttl & 0x00FF0000) >> 16;
+	if (client->ednsversion > 0) {
+		isc_stats_increment(ns_g_server->nsstats,
+				    dns_nsstatscounter_badednsver);
+		result = ns_client_addopt(client, client->message,
+					  &client->opt);
+		if (result == ISC_R_SUCCESS)
+			result = DNS_R_BADVERS;
+		ns_client_error(client, result);
+		goto cleanup;
+	}
+
+	/* Check for NSID request */
+	result = dns_rdataset_first(opt);
+	if (result == ISC_R_SUCCESS) {
+		dns_rdata_init(&rdata);
+		dns_rdataset_current(opt, &rdata);
+		isc_buffer_init(&optbuf, rdata.data, rdata.length);
+		isc_buffer_add(&optbuf, rdata.length);
+		while (isc_buffer_remaininglength(&optbuf) >= 4) {
+			optcode = isc_buffer_getuint16(&optbuf);
+			optlen = isc_buffer_getuint16(&optbuf);
+			switch (optcode) {
+			case DNS_OPT_NSID:
+				if (!WANTNSID(client))
+					isc_stats_increment(
+						    ns_g_server->nsstats,
+						    dns_nsstatscounter_nsidopt);
+				client->attributes |= NS_CLIENTATTR_WANTNSID;
+				isc_buffer_forward(&optbuf, optlen);
+				break;
+#ifdef ISC_PLATFORM_USESIT
+			case DNS_OPT_COOKIE:
+				process_cookie(client, &optbuf, optlen);
+				break;
+#endif
+			case DNS_OPT_EXPIRE:
+				if (!WANTEXPIRE(client))
+					isc_stats_increment(
+						  ns_g_server->nsstats,
+						  dns_nsstatscounter_expireopt);
+				client->attributes |= NS_CLIENTATTR_WANTEXPIRE;
+				isc_buffer_forward(&optbuf, optlen);
+				break;
+			default:
+				isc_stats_increment(ns_g_server->nsstats,
+						  dns_nsstatscounter_otheropt);
+				isc_buffer_forward(&optbuf, optlen);
+				break;
+			}
+		}
+	}
+
+	isc_stats_increment(ns_g_server->nsstats, dns_nsstatscounter_edns0in);
+	client->attributes |= NS_CLIENTATTR_WANTOPT;
+
+ cleanup:
+	return (result);
+}
+
 /*
  * Handle an incoming request event from the socket (UDP case)
  * or tcpmsg (TCP case).
@@ -1379,8 +1855,6 @@ client_request(isc_task_t *task, isc_event_t *event) {
 	dns_messageid_t id;
 	unsigned int flags;
 	isc_boolean_t notimp;
-	dns_rdata_t rdata;
-	isc_uint16_t optcode;
 
 	REQUIRE(event != NULL);
 	client = event->ev_arg;
@@ -1389,9 +1863,9 @@ client_request(isc_task_t *task, isc_event_t *event) {
 
 	INSIST(client->recursionquota == NULL);
 
-	INSIST(client->state == TCP_CLIENT(client) ?
+	INSIST(client->state == (TCP_CLIENT(client) ?
 				       NS_CLIENTSTATE_READING :
-				       NS_CLIENTSTATE_READY);
+				       NS_CLIENTSTATE_READY));
 
 	ns_client_requests++;
 
@@ -1407,6 +1881,13 @@ client_request(isc_task_t *task, isc_event_t *event) {
 			client->peeraddr = sevent->address;
 			client->peeraddr_valid = ISC_TRUE;
 		}
+		if ((sevent->attributes & ISC_SOCKEVENTATTR_DSCP) != 0) {
+			ns_client_log(client, NS_LOGCATEGORY_CLIENT,
+			      NS_LOGMODULE_CLIENT, ISC_LOG_DEBUG(90),
+			      "received DSCP %d", sevent->dscp);
+			if (client->dscp == -1)
+				client->dscp = sevent->dscp;
+		}
 		if ((sevent->attributes & ISC_SOCKEVENTATTR_PKTINFO) != 0) {
 			client->attributes |= NS_CLIENTATTR_PKTINFO;
 			client->pktinfo = sevent->pktinfo;
@@ -1540,7 +2021,7 @@ client_request(isc_task_t *task, isc_event_t *event) {
 	}
 	if (TCP_CLIENT(client))
 		isc_stats_increment(ns_g_server->nsstats,
-				    dns_nsstatscounter_tcp);
+				    dns_nsstatscounter_requesttcp);
 
 	/*
 	 * It's a request.  Parse it.
@@ -1551,6 +2032,14 @@ client_request(isc_task_t *task, isc_event_t *event) {
 		 * Parsing the request failed.  Send a response
 		 * (typically FORMERR or SERVFAIL).
 		 */
+		if (result == DNS_R_OPTERR)
+			(void)ns_client_addopt(client, client->message,
+					       &client->opt);
+
+		ns_client_log(client, NS_LOGCATEGORY_CLIENT,
+			      NS_LOGMODULE_CLIENT, ISC_LOG_DEBUG(1),
+			      "message parsing failed: %s",
+			      isc_result_totext(result));
 		ns_client_error(client, result);
 		goto cleanup;
 	}
@@ -1578,69 +2067,21 @@ client_request(isc_task_t *task, isc_event_t *event) {
 	/*
 	 * Deal with EDNS.
 	 */
-	opt = dns_message_getopt(client->message);
+	if (ns_g_noedns)
+		opt = NULL;
+	else
+		opt = dns_message_getopt(client->message);
 	if (opt != NULL) {
 		/*
-		 * Set the client's UDP buffer size.
+		 * Are we dropping all EDNS queries?
 		 */
-		client->udpsize = opt->rdclass;
-
-		/*
-		 * If the requested UDP buffer size is less than 512,
-		 * ignore it and use 512.
-		 */
-		if (client->udpsize < 512)
-			client->udpsize = 512;
-
-		/*
-		 * Get the flags out of the OPT record.
-		 */
-		client->extflags = (isc_uint16_t)(opt->ttl & 0xFFFF);
-
-		/*
-		 * Do we understand this version of EDNS?
-		 *
-		 * XXXRTH need library support for this!
-		 */
-		client->ednsversion = (opt->ttl & 0x00FF0000) >> 16;
-		if (client->ednsversion > 0) {
-			isc_stats_increment(ns_g_server->nsstats,
-					    dns_nsstatscounter_badednsver);
-			result = client_addopt(client);
-			if (result == ISC_R_SUCCESS)
-				result = DNS_R_BADVERS;
-			ns_client_error(client, result);
+		if (ns_g_dropedns) {
+			ns_client_next(client, ISC_R_SUCCESS);
 			goto cleanup;
 		}
-
-		/* Check for NSID request */
-		result = dns_rdataset_first(opt);
-		if (result == ISC_R_SUCCESS) {
-			dns_rdata_init(&rdata);
-			dns_rdataset_current(opt, &rdata);
-			if (rdata.length >= 2) {
-				isc_buffer_t nsidbuf;
-				isc_buffer_init(&nsidbuf,
-						rdata.data, rdata.length);
-				isc_buffer_add(&nsidbuf, rdata.length);
-				optcode = isc_buffer_getuint16(&nsidbuf);
-				if (optcode == DNS_OPT_NSID)
-					client->attributes |=
-						NS_CLIENTATTR_WANTNSID;
-			}
-		}
-
-		isc_stats_increment(ns_g_server->nsstats,
-				    dns_nsstatscounter_edns0in);
-
-		/*
-		 * Create an OPT for our reply.
-		 */
-		result = client_addopt(client);
-		if (result != ISC_R_SUCCESS) {
-			ns_client_error(client, result);
+		result = process_opt(client, opt);
+		if (result != ISC_R_SUCCESS)
 			goto cleanup;
-		}
 	}
 
 	if (client->message->rdclass == 0) {
@@ -2050,6 +2491,8 @@ client_create(ns_clientmgr_t *manager, ns_client_t **clientp) {
 		goto cleanup_task;
 	client->timerset = ISC_FALSE;
 
+	client->delaytimer = NULL;
+
 	client->message = NULL;
 	result = dns_message_create(client->mctx, DNS_MESSAGE_INTENTPARSE,
 				    &client->message);
@@ -2058,11 +2501,9 @@ client_create(ns_clientmgr_t *manager, ns_client_t **clientp) {
 
 	/* XXXRTH  Hardwired constants */
 
-	client->sendevent = (isc_socketevent_t *)
-			    isc_event_allocate(client->mctx, client,
-					       ISC_SOCKEVENT_SENDDONE,
-					       client_senddone, client,
-					       sizeof(isc_socketevent_t));
+	client->sendevent = isc_socket_socketevent(client->mctx, client,
+						   ISC_SOCKEVENT_SENDDONE,
+						   client_senddone, client);
 	if (client->sendevent == NULL) {
 		result = ISC_R_NOMEMORY;
 		goto cleanup_message;
@@ -2074,11 +2515,9 @@ client_create(ns_clientmgr_t *manager, ns_client_t **clientp) {
 		goto cleanup_sendevent;
 	}
 
-	client->recvevent = (isc_socketevent_t *)
-			    isc_event_allocate(client->mctx, client,
-					       ISC_SOCKEVENT_RECVDONE,
-					       client_request, client,
-					       sizeof(isc_socketevent_t));
+	client->recvevent = isc_socket_socketevent(client->mctx, client,
+						   ISC_SOCKEVENT_RECVDONE,
+						   client_request, client);
 	if (client->recvevent == NULL) {
 		result = ISC_R_NOMEMORY;
 		goto cleanup_recvbuf;
@@ -2105,6 +2544,7 @@ client_create(ns_clientmgr_t *manager, ns_client_t **clientp) {
 	client->tcpbuf = NULL;
 	client->opt = NULL;
 	client->udpsize = 512;
+	client->dscp = -1;
 	client->extflags = 0;
 	client->ednsversion = -1;
 	client->next = NULL;
@@ -2117,9 +2557,11 @@ client_create(ns_clientmgr_t *manager, ns_client_t **clientp) {
 	client->recursionquota = NULL;
 	client->interface = NULL;
 	client->peeraddr_valid = ISC_FALSE;
-#ifdef ALLOW_FILTER_AAAA_ON_V4
-	client->filter_aaaa = dns_v4_aaaa_ok;
+#ifdef ALLOW_FILTER_AAAA
+	client->filter_aaaa = dns_aaaa_ok;
 #endif
+	client->needshutdown = ns_g_clienttest;
+
 	ISC_EVENT_INIT(&client->ctlevent, sizeof(client->ctlevent), 0, NULL,
 		       NS_EVENT_CLIENTCONTROL, client_start, client, client,
 		       NULL, NULL);
@@ -2147,8 +2589,6 @@ client_create(ns_clientmgr_t *manager, ns_client_t **clientp) {
 	if (result != ISC_R_SUCCESS)
 		goto cleanup_query;
 
-	client->needshutdown = ns_g_clienttest;
-
 	CTRACE("create");
 
 	*clientp = client;
@@ -2410,6 +2850,9 @@ ns_client_replace(ns_client_t *client) {
 
 	CTRACE("replace");
 
+	REQUIRE(client != NULL);
+	REQUIRE(client->manager != NULL);
+
 	result = get_client(client->manager, client->interface,
 			    client->dispatch, TCP_CLIENT(client));
 	if (result != ISC_R_SUCCESS)
@@ -2501,10 +2944,10 @@ ns_clientmgr_create(isc_mem_t *mctx, isc_taskmgr_t *taskmgr,
 	return (ISC_R_SUCCESS);
 
  cleanup_listlock:
-	isc_mutex_destroy(&manager->listlock);
+	(void) isc_mutex_destroy(&manager->listlock);
 
  cleanup_lock:
-	isc_mutex_destroy(&manager->lock);
+	(void) isc_mutex_destroy(&manager->lock);
 
  cleanup_manager:
 	isc_mem_put(manager->mctx, manager, sizeof(*manager));
@@ -2514,9 +2957,10 @@ ns_clientmgr_create(isc_mem_t *mctx, isc_taskmgr_t *taskmgr,
 
 void
 ns_clientmgr_destroy(ns_clientmgr_t **managerp) {
+	isc_result_t result;
 	ns_clientmgr_t *manager;
 	ns_client_t *client;
-	isc_boolean_t need_destroy = ISC_FALSE;
+	isc_boolean_t need_destroy = ISC_FALSE, unlock = ISC_FALSE;
 
 	REQUIRE(managerp != NULL);
 	manager = *managerp;
@@ -2524,11 +2968,16 @@ ns_clientmgr_destroy(ns_clientmgr_t **managerp) {
 
 	MTRACE("destroy");
 
-	LOCK(&manager->listlock);
+	/*
+	 * Check for success because we may already be task-exclusive
+	 * at this point.  Only if we succeed at obtaining an exclusive
+	 * lock now will we need to relinquish it later.
+	 */
+	result = isc_task_beginexclusive(ns_g_server->task);
+	if (result == ISC_R_SUCCESS)
+		unlock = ISC_TRUE;
 
-	LOCK(&manager->lock);
 	manager->exiting = ISC_TRUE;
-	UNLOCK(&manager->lock);
 
 	for (client = ISC_LIST_HEAD(manager->clients);
 	     client != NULL;
@@ -2538,7 +2987,8 @@ ns_clientmgr_destroy(ns_clientmgr_t **managerp) {
 	if (ISC_LIST_EMPTY(manager->clients))
 		need_destroy = ISC_TRUE;
 
-	UNLOCK(&manager->listlock);
+	if (unlock)
+		isc_task_endexclusive(ns_g_server->task);
 
 	if (need_destroy)
 		clientmgr_destroy(manager);
@@ -2555,6 +3005,11 @@ get_client(ns_clientmgr_t *manager, ns_interface_t *ifp,
 	ns_client_t *client;
 	MTRACE("get client");
 
+	REQUIRE(manager != NULL);
+
+	if (manager->exiting)
+		return (ISC_R_SHUTTINGDOWN);
+
 	/*
 	 * Allocate a client.  First try to get a recycled one;
 	 * if that fails, make a new one.
@@ -2584,6 +3039,8 @@ get_client(ns_clientmgr_t *manager, ns_interface_t *ifp,
 	client->state = NS_CLIENTSTATE_READY;
 	INSIST(client->recursionquota == NULL);
 
+	client->dscp = ifp->dscp;
+
 	if (tcp) {
 		client->attributes |= NS_CLIENTATTR_TCP;
 		isc_socket_attach(ifp->tcpsocket,
@@ -2694,7 +3151,8 @@ ns_client_checkacl(ns_client_t *client, isc_sockaddr_t *sockaddr,
 static void
 ns_client_name(ns_client_t *client, char *peerbuf, size_t len) {
 	if (client->peeraddr_valid)
-		isc_sockaddr_format(&client->peeraddr, peerbuf, len);
+		isc_sockaddr_format(&client->peeraddr, peerbuf,
+				    (unsigned int)len);
 	else
 		snprintf(peerbuf, len, "@%p", client);
 }
@@ -2703,7 +3161,7 @@ void
 ns_client_logv(ns_client_t *client, isc_logcategory_t *category,
 	       isc_logmodule_t *module, int level, const char *fmt, va_list ap)
 {
-	char msgbuf[2048];
+	char msgbuf[4096];
 	char peerbuf[ISC_SOCKADDR_FORMATSIZE];
 	char signerbuf[DNS_NAME_FORMATSIZE], qnamebuf[DNS_NAME_FORMATSIZE];
 	const char *viewname = "";
@@ -2778,6 +3236,9 @@ ns_client_dumpmessage(ns_client_t *client, const char *reason) {
 	int len = 1024;
 	isc_result_t result;
 
+	if (!isc_log_wouldlog(ns_g_lctx, ISC_LOG_DEBUG(1)))
+		return;
+
 	/*
 	 * Note that these are multiline debug messages.  We want a newline
 	 * to appear in the log after each message.
